/*
 * SparseGraph PostgreSQL extension
 * Copyright (C) 2007 Antal Novak
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "sparsegraph.h"

#include "funcapi.h"
#include "catalog/pg_type.h"
#include "access/heapam.h"
#include "utils/array.h"
#include "access/tuptoaster.h"

PG_FUNCTION_INFO_V1( sparsegraph_in );
Datum
sparsegraph_in( PG_FUNCTION_ARGS )
{
	const char *str = PG_GETARG_CSTRING( 0 );
	struct sparsegraph *result = sparsegraph_import( str );
	PG_RETURN_POINTER( result );
}

PG_FUNCTION_INFO_V1( sparsegraph_out );
Datum
sparsegraph_out( PG_FUNCTION_ARGS )
{
	struct sparsegraph *mtx = (struct sparsegraph *) PG_GETARG_VARLENA_P( 0 );
	const char *result = sparsegraph_export( mtx );
	PG_RETURN_CSTRING( result );
}

PG_FUNCTION_INFO_V1( sparsegraph_neighbors );
Datum
sparsegraph_neighbors( PG_FUNCTION_ARGS )
{
	FuncCallContext *ctx;
	Datum values[3];
	bool nulls[3];
	HeapTuple tuple;

	struct my_context
	{
		struct sparsegraph *mtx;
		struct varlena *mtx_varlena;
		union sparsegraph_block *edge_cur;
		union sparsegraph_block *edge_end;
		int32 *label_next;
		int32 *label_end;
	} *myctx = 0;

	if( SRF_IS_FIRSTCALL() )
	{
		ArrayType *labels;
		MemoryContext oldcontext;
		TupleDesc tuple_desc;

		ctx = SRF_FIRSTCALL_INIT();
		oldcontext = MemoryContextSwitchTo( ctx->multi_call_memory_ctx );

		ctx->user_fctx = myctx = (struct my_context *) palloc( sizeof( struct my_context ) );
		MemSet( myctx, 0, sizeof( *myctx ) );

		/* Read in the arguments */
		myctx->mtx_varlena = PG_GETARG_RAW_VARLENA_P( 0 );
		labels = PG_GETARG_ARRAYTYPE_P( 1 );

		/* Make sure the label set is valid */
		if( ARR_NDIM( labels ) != 1 || ARR_ELEMTYPE( labels ) != INT4OID )
		{
			ereport( ERROR,
			         (errcode(ERRCODE_ARRAY_ELEMENT_ERROR),
		              errmsg( "sparsegraph_neighbors(): Expected an array of INTs" ) ) );
		}

		myctx->label_next  = (int32 *) ARR_DATA_PTR( labels );
		myctx->label_end   = myctx->label_next + ARR_DIMS(labels)[0];

		/* Set up the output TupleDesc */
		tuple_desc = CreateTemplateTupleDesc( 3, false );
		TupleDescInitEntry( tuple_desc, (AttrNumber) 1, "label1", INT4OID,   -1, 0 );
		TupleDescInitEntry( tuple_desc, (AttrNumber) 2, "label2", INT4OID,   -1, 0 );
		TupleDescInitEntry( tuple_desc, (AttrNumber) 3, "weight", FLOAT8OID, -1, 0 );
		ctx->tuple_desc = BlessTupleDesc( tuple_desc );

		/* De-toast enough of the sparsegraph struct to get the nodes.
		 * We first de-toast enough to get num_nodes... */
		myctx->mtx = (struct sparsegraph *)
			pg_detoast_datum_slice( myctx->mtx_varlena, 0, sizeof( struct sparsegraph ) - VARHDRSZ );

		/* Have to check the version to make sure we're compatible */
		if( myctx->mtx->version > SPARSEGRAPH_CURRENT )
		{
			ereport( ERROR,
			         (errcode(ERRCODE_DATATYPE_MISMATCH),
			          errmsg( "sparsegraph code version (0x%08x) is older than data "
			                  "version (0x%08x)!", SPARSEGRAPH_CURRENT, myctx->mtx->version ) ) );
		}

		/* Then --- once we have the size of the node header --- de-toast the
		 * full node header.
		 */
		myctx->mtx = (struct sparsegraph *)
			pg_detoast_datum_slice( myctx->mtx_varlena, 0, sizeof( struct sparsegraph )
			                                               + sizeof( union sparsegraph_block ) * myctx->mtx->num_nodes
			                                               - VARHDRSZ );

		MemoryContextSwitchTo( oldcontext );
	}

	ctx   = SRF_PERCALL_SETUP();
	myctx = ctx->user_fctx;

	while( 1 )
	{
		if( myctx->edge_cur < myctx->edge_end )
		{
			values[0] = Int32GetDatum( *( myctx->label_next - 1 ) );
			values[1] = Int32GetDatum( myctx->mtx->data[myctx->edge_cur->node_idx].label );
			values[2] = Float8GetDatum( myctx->edge_cur->weight );
			++myctx->edge_cur;

			MemSet( nulls, 0, sizeof(nulls) );
			tuple = heap_form_tuple( ctx->tuple_desc, values, nulls );
			SRF_RETURN_NEXT( ctx, HeapTupleGetDatum( tuple ) );
		}
		else
		{
			/* Advance the node pointer, if we haven't gone too far already */
			if( myctx->label_next < myctx->label_end )
			{
				MemoryContext oldcontext = MemoryContextSwitchTo( ctx->multi_call_memory_ctx );
				int32 label = *myctx->label_next;
				int32 idx = myctx->mtx->num_nodes;

				/* labels are sorted, so we'll binary search for 'label' */
				{
					int32 low = 0;
					int32 high = myctx->mtx->num_nodes;

					while( low < high )
					{
						/* pi = probe index, pv = probe value */
						int32 pi   = low + (high-low)/2;
						label_t pv = myctx->mtx->data[pi].label;

						if( label < pv )
							high = pi;
						else if( label > pv )
							low = pi + 1;
						else
						{
							idx = pi;
							break;
						}
					}
				}

				if( idx < myctx->mtx->num_nodes )
				{
					myctx->edge_cur = (union sparsegraph_block *)
						pg_detoast_datum_slice( myctx->mtx_varlena,
												sizeof( struct sparsegraph )
												+ sizeof( union sparsegraph_block ) * myctx->mtx->data[idx].data_idx
												- VARHDRSZ,
												sizeof( union sparsegraph_block ) * myctx->mtx->data[idx].num_edges );

					/* Skip VARHDRSZ bytes */
					myctx->edge_cur = (union sparsegraph_block *) (((char *) myctx->edge_cur) + VARHDRSZ);
					myctx->edge_end = myctx->edge_cur + myctx->mtx->data[idx].num_edges;
				}
				else
				{
					myctx->edge_cur = myctx->edge_end = 0;
				}

				MemoryContextSwitchTo( oldcontext );

				++myctx->label_next;
			}
			else
			{
				/* Otherwise, we're done! */
				SRF_RETURN_DONE( ctx );
			}
		}
	}
}

PG_FUNCTION_INFO_V1( sparsegraph_stats );
Datum
sparsegraph_stats( PG_FUNCTION_ARGS )
{
	Datum values[2];
	bool nulls[2];
	HeapTuple tuple;
	TupleDesc tuple_desc;

	/* Only de-toast enough of the sparsegraph struct to get the node count */
	struct varlena *mtx_varlena = PG_GETARG_RAW_VARLENA_P( 0 );
	int32 rawsize = toast_raw_datum_size( PG_GETARG_DATUM( 0 ) );

	struct sparsegraph *mtx = (struct sparsegraph *)
		pg_detoast_datum_slice( mtx_varlena, 0, sizeof( struct sparsegraph ) - VARHDRSZ );

	/* Have to check the version to make sure we're compatible */
	if( mtx->version > SPARSEGRAPH_CURRENT )
	{
		ereport( ERROR,
		         (errcode(ERRCODE_DATATYPE_MISMATCH),
		          errmsg( "sparsegraph code version (0x%08x) is older than data "
		                  "version (0x%08x)!", SPARSEGRAPH_CURRENT, mtx->version ) ) );
	}

	tuple_desc = CreateTemplateTupleDesc( 2, false );
	TupleDescInitEntry( tuple_desc, (AttrNumber) 1, "node_count", INT4OID,   -1, 0 );
	TupleDescInitEntry( tuple_desc, (AttrNumber) 2, "edge_count", INT4OID,   -1, 0 );
	tuple_desc = BlessTupleDesc( tuple_desc );

	values[0] = Int32GetDatum( mtx->num_nodes );
	values[1] = Int32GetDatum( (rawsize - sizeof(*mtx)) / sizeof(union sparsegraph_block) - mtx->num_nodes );

	MemSet( nulls, 0, sizeof(nulls) );
	tuple = heap_form_tuple( tuple_desc, values, nulls );
	return HeapTupleGetDatum( tuple );
}

PG_FUNCTION_INFO_V1( sparsegraph_nodes );
Datum
sparsegraph_nodes( PG_FUNCTION_ARGS )
{
	FuncCallContext *ctx;

	struct my_context
	{
		union sparsegraph_block *node_cur;
		union sparsegraph_block *node_end;
	} *myctx = 0;

	if( SRF_IS_FIRSTCALL() )
	{
		MemoryContext oldcontext;
		struct varlena *mtx_varlena;
		struct sparsegraph *mtx;

		ctx = SRF_FIRSTCALL_INIT();
		oldcontext = MemoryContextSwitchTo( ctx->multi_call_memory_ctx );

		ctx->user_fctx = myctx = (struct my_context *) palloc( sizeof( struct my_context ) );
		MemSet( myctx, 0, sizeof( *myctx ) );

		/* Read in the arguments */
		mtx_varlena = PG_GETARG_RAW_VARLENA_P( 0 );

		/* De-toast enough of the sparsegraph struct to get the nodes.
		 * We first de-toast enough to get num_nodes... */
		mtx = (struct sparsegraph *)
			pg_detoast_datum_slice( mtx_varlena, 0, sizeof( struct sparsegraph ) - VARHDRSZ );

		/* Have to check the version to make sure we're compatible */
		if( mtx->version > SPARSEGRAPH_CURRENT )
		{
			ereport( ERROR,
			         (errcode(ERRCODE_DATATYPE_MISMATCH),
			          errmsg( "sparsegraph code version (0x%08x) is older than data "
			                  "version (0x%08x)!", SPARSEGRAPH_CURRENT, mtx->version ) ) );
		}

		/* Then --- once we have the size of the node header --- de-toast the
		 * full node header.
		 */
		mtx = (struct sparsegraph *)
			pg_detoast_datum_slice( mtx_varlena, 0, sizeof( struct sparsegraph )
			                                        + sizeof( union sparsegraph_block ) * mtx->num_nodes
			                                        - VARHDRSZ );

		myctx->node_cur = &mtx->data[0];
		myctx->node_end = myctx->node_cur + mtx->num_nodes;

		MemoryContextSwitchTo( oldcontext );
	}

	ctx   = SRF_PERCALL_SETUP();
	myctx = ctx->user_fctx;

	/* Any nodes left? */
	if( myctx->node_cur < myctx->node_end )
	{
		int32 label = myctx->node_cur->label;
		++myctx->node_cur;

		SRF_RETURN_NEXT( ctx, Int32GetDatum( label ) );
	}
	else
	{
		/* Otherwise, we're done! */
		SRF_RETURN_DONE( ctx );
	}
}

PG_FUNCTION_INFO_V1( sparsegraph_edges );
Datum
sparsegraph_edges( PG_FUNCTION_ARGS )
{
	FuncCallContext *ctx;
	Datum values[3];
	bool nulls[3];
	HeapTuple tuple;

	struct my_context
	{
		struct sparsegraph *mtx;
		union sparsegraph_block *edge_cur;
		union sparsegraph_block *edge_end;
		union sparsegraph_block *node_next;
		union sparsegraph_block *node_end;
		label_t label_cur;
	} *myctx = 0;

	if( SRF_IS_FIRSTCALL() )
	{
		MemoryContext oldcontext;
		TupleDesc tuple_desc;

		ctx = SRF_FIRSTCALL_INIT();
		oldcontext = MemoryContextSwitchTo( ctx->multi_call_memory_ctx );

		ctx->user_fctx = myctx = (struct my_context *) palloc( sizeof( struct my_context ) );
		MemSet( myctx, 0, sizeof( *myctx ) );

		/* Read in the arguments -- we need to completely de-toast the sparsegraph in this case */
		myctx->mtx = (struct sparsegraph *) PG_GETARG_VARLENA_P( 0 );

		/* Have to check the version to make sure we're compatible */
		if( myctx->mtx->version > SPARSEGRAPH_CURRENT )
		{
			ereport( ERROR,
			         (errcode(ERRCODE_DATATYPE_MISMATCH),
			          errmsg( "sparsegraph code version (0x%08x) is older than data "
			                  "version (0x%08x)!", SPARSEGRAPH_CURRENT, myctx->mtx->version ) ) );
		}

		myctx->node_next  = &myctx->mtx->data[0];
		myctx->node_end   = myctx->node_next + myctx->mtx->num_nodes;

		/* Set up the output TupleDesc */
		tuple_desc = CreateTemplateTupleDesc( 3, false );
		TupleDescInitEntry( tuple_desc, (AttrNumber) 1, "label1", INT4OID,   -1, 0 );
		TupleDescInitEntry( tuple_desc, (AttrNumber) 2, "label2", INT4OID,   -1, 0 );
		TupleDescInitEntry( tuple_desc, (AttrNumber) 3, "weight", FLOAT8OID, -1, 0 );
		ctx->tuple_desc = BlessTupleDesc( tuple_desc );

		MemoryContextSwitchTo( oldcontext );
	}

	ctx   = SRF_PERCALL_SETUP();
	myctx = ctx->user_fctx;

	while( 1 )
	{
		if( myctx->edge_cur < myctx->edge_end )
		{
			values[0] = Int32GetDatum( myctx->label_cur );
			values[1] = Int32GetDatum( myctx->mtx->data[myctx->edge_cur->node_idx].label );
			values[2] = Float8GetDatum( myctx->edge_cur->weight );
			++myctx->edge_cur;

			MemSet( nulls, 0, sizeof(nulls) );
			tuple = heap_form_tuple( ctx->tuple_desc, values, nulls );
			SRF_RETURN_NEXT( ctx, HeapTupleGetDatum( tuple ) );
		}
		else
		{
			/* Advance the node pointer, if we haven't gone too far already */
			if( myctx->node_next < myctx->node_end )
			{
				myctx->label_cur = myctx->node_next->label;
				myctx->edge_cur  = &myctx->mtx->data[myctx->node_next->data_idx];
				myctx->edge_end  = myctx->edge_cur + myctx->node_next->num_edges;

				++myctx->node_next;
			}
			else
			{
				/* Otherwise, we're done! */
				SRF_RETURN_DONE( ctx );
			}
		}
	}
}
