/*
 * SparseGraph PostgreSQL extension
 * Copyright (C) 2007 Antal Novak
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef _SPARSE_MATRIX_H_
#define _SPARSE_MATRIX_H_

#ifdef __cplusplus
extern "C" {
#endif

#ifndef NO_POSTGRESQL
#  include "postgres.h"
#  include "fmgr.h"
#else /* NO_POSTGRESQL */
#  define int32 int
#  define float8 double
#  define palloc malloc
#  define ereport(x,y) do { y; printf("\n"); exit(1); } while(0)
#  define errcode(x) (void)0
#  define errmsg printf
#endif /* NO_POSTGRESQL */

/** The current version number (see sparsegraph.version) */
#define SPARSEGRAPH_CURRENT 0x00010001

/** Typedef for node labels */
typedef int32 label_t;

/** Typedef for edge labels (weights) */
typedef float8 weight_t;

/**
 * Compact representation of a sparse matrix with int32 labels and float8
 * values.
 *
 * Header consists of:
 *   - size (total size, for varlena compatibility)
 *   - num_nodes (number of nodes in the graph)
 *   - one sparsegraph_block for each node, sorted by label
 *   - one sparsegraph_block for each edge
 */
struct sparsegraph
{
	/* the total size of the data structure in bytes */
	int32 size;

	/* A version number, so we can safely change the internal structure later.
	 * By convention, this will be of the form 0x11112222, where 1111 is the
	 * major version and 2222 is the minor.
	 */
	int32 version;

	/* number of nodes in the graph */
	int32 num_nodes;

	union sparsegraph_block
	{
		/* the data[] array contains num_nodes of these (node) structs */
		struct
		{
			int32 data_idx;
			label_t label;
			int32 num_edges;
		};

		/* followed by one (edge) struct for each edge */
		struct
		{
			int32 node_idx;
			weight_t weight;
		};
	} data[0];
};

/**
 * Allocate a new sparsegraph struct with 'n' nodes and 'm' edges
 */
struct sparsegraph *sparsegraph_alloc(int n, int m);

/**
 * Import a sparsegraph from the string representation, which is of the form:
 *
 * [Prefixes] Elements
 *
 * Prefixes   := <empty> | ":" PrefixList ":"
 * PrefixList := Prefix | Prefix ";" PrefixList
 * Prefix     := "undirected"
 *
 * Elements := <empty> | Element [","] Elements
 * Element  := "(" NodeLabel ")"                               # declare a node (no edges)
 *           | "(" NodeLabel "," NodeLabel "," EdgeWeight ")"  # declare an edge
 */
struct sparsegraph *sparsegraph_import(const char *str);

/**
 * Export a sparsegraph into the external string representation.
 */
const char *sparsegraph_export(struct sparsegraph *mtx);

#ifndef NO_POSTGRESQL

/**
 * Wrapper for sparsegraph_import()
 */
Datum sparsegraph_in( PG_FUNCTION_ARGS );

/**
 * Wrapper for sparsegraph_export()
 */
Datum sparsegraph_out( PG_FUNCTION_ARGS );

/**
 * Given a sparsegraph and an array of labels (INT4), return all edges
 * incident to any of these labels.  Returns a set of records
 * (label1 INT4, label2 INT4, weight FLOAT8).
 */
Datum sparsegraph_neighbors( PG_FUNCTION_ARGS );

/**
 * Given a sparse matrix, return a two-column record
 * (node_count INT4, edge_count INT4).
 */
Datum sparsegraph_stats( PG_FUNCTION_ARGS );

/**
 * Given a sparse matrix, get all nodes.  Returns a set of INT4.
 */
Datum sparsegraph_nodes( PG_FUNCTION_ARGS );

/**
 * Given a sparse matrix, get all edges.  Returns a set of records
 * (label1 INT4, label2 INT4, weight FLOAT8).
 */
Datum sparsegraph_edges( PG_FUNCTION_ARGS );

#endif /* !NO_POSTGRESQL */

#ifdef __cplusplus
}
#endif

#endif /* _SPARSE_MATRIX_H_ */
