/*
 * SparseGraph PostgreSQL extension
 * Copyright (C) 2007 Antal Novak
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "sparsegraph.h"
#include <map>
#include <set>
#include <vector>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;

static
const char *pstring( const string &s )
{
	char *result = (char *) palloc( s.size() + 1 );
	strncpy( result, s.c_str(), s.size() + 1 );
	return result;
}

static
string truncate_str( const string &s, size_t n )
{
	if( s.size() <= n )
	{
		return s;
	}

	if( n > 3 )
	{
		return s.substr( 0, n-3 ) + "...";
	}

	return s.substr( 0, n );
}

static
const char *truncate_cstr( const char *s, size_t n )
{
	return pstring( truncate_str( s, n ) );
}

/**
 * Do something like:
 *
 *    sscanf( " ( %d , %d , %lf ) %n", ... );
 *
 * but MUCH faster :-)
 *
 * Returns number of bytes read on a complete successful parse, zero
 * otherwise.
 */
static size_t
sscan_edge( const char *const str, int *a, int *b, double *c )
{
	register const char *s;
	register char *endptr;

	s = str;

	while(*s == ' ') ++s;
	if(*s++ != '(') return 0;
	while(*s == ' ') ++s;

	*a = strtol(s, &endptr, 10);
	if(endptr == s) return 0;
	s = endptr;

	while(*s == ' ') ++s;
	if(*s++ != ',') return 0;
	while(*s == ' ') ++s;

	*b = strtol(s, &endptr, 10);
	if(endptr == s) return 0;
	s = endptr;

	while(*s == ' ') ++s;
	if(*s++ != ',') return 0;
	while(*s == ' ') ++s;

	*c = strtod(s, &endptr);
	if(endptr == s) return 0;
	s = endptr;

	while(*s == ' ') ++s;
	if(*s++ != ')') return 0;
	while(*s == ' ') ++s;

	return s - str;
}

/**
 * Do something like:
 *
 *    sscanf( " ( %d ) %n", ... );
 *
 * but MUCH faster :-)
 *
 * Returns number of bytes read on a complete successful parse, zero
 * otherwise.
 */
static size_t
sscan_node( const char *const str, int *a )
{
	register const char *s;
	register char *endptr;

	s = str;

	while(*s == ' ') ++s;
	if(*s++ != '(') return 0;
	while(*s == ' ') ++s;

	*a = strtol(s, &endptr, 10);
	if(endptr == s) return 0;
	s = endptr;

	while(*s == ' ') ++s;
	if(*s++ != ')') return 0;
	while(*s == ' ') ++s;

	return s - str;
}

class FatMatrix
{
public:
	void addEdge( label_t __l1, label_t __l2, weight_t __weight )
	{
		addNode( __l1 );
		addNode( __l2 );
		_edges[make_pair(__l1,__l2)] = __weight;
	}

	void addNode( label_t label )
	{
		_labels.insert( label );
	}

	struct sparsegraph *newSparseMatrix()
	{
		/** Convert our _labels and _edges to use indexes rather than labels */
		typedef vector<size_t> INodeLabels;
		typedef map<label_t, size_t> INodeIndexes;
		typedef map< pair<size_t,size_t>, weight_t > IEdges;

		INodeLabels labels;
		INodeIndexes indexes;
		IEdges edges;

		for( NodeLabels::const_iterator it = _labels.begin(), end = _labels.end() ;
		     it != end ;
		     ++it )
		{
			indexes[*it] = labels.size();
			labels.push_back( *it );
		}

		for( Edges::const_iterator it = _edges.begin(), end = _edges.end() ;
		     it != end ;
		     ++it )
		{
			edges[make_pair( indexes[it->first.first], indexes[it->first.second] )] = it->second;
		}

		int32 n = labels.size();
		int32 m = edges.size();

		struct sparsegraph *mtx = sparsegraph_alloc(n, m);

		size_t cur_offset = n;

		for( int32 i = 0 ; i < n ; ++i )
		{
			IEdges::iterator lower = edges.lower_bound( make_pair( i, 0 ) );
			IEdges::iterator upper = edges.lower_bound( make_pair( i+1, 0 ) );

			int32 num_edges = distance( lower, upper );

			mtx->data[i].data_idx  = cur_offset;
			mtx->data[i].num_edges = num_edges;
			mtx->data[i].label     = labels[i];

			for( ; lower != upper ; ++lower, ++cur_offset )
			{
				mtx->data[cur_offset].node_idx = lower->first.second;
				mtx->data[cur_offset].weight   = lower->second;
			}
		}

		return mtx;
	}

private:
	/** The set of node labels */
	typedef set<label_t> NodeLabels;

	/** Map from a pair of node labels to an edge weight */
	typedef map< pair<label_t,label_t>, weight_t > Edges;

	NodeLabels _labels;
	Edges _edges;
};

struct sparsegraph *
sparsegraph_alloc(int n, int m)
{
	struct sparsegraph *mtx = 0;

	int32 size = sizeof(*mtx) +
	             n * sizeof(*mtx->data) +
	             m * sizeof(*mtx->data);

	mtx = (struct sparsegraph *) palloc( size );
	bzero( mtx, size );

	mtx->size      = size;
	mtx->version   = SPARSEGRAPH_CURRENT;
	mtx->num_nodes = n;

	return mtx;
}

struct sparsegraph *sparsegraph_import(const char *str)
{
	FatMatrix mtx;
	bool is_undirected = false;
	int bytes_read;
	int result;

	/* Parse prefixes */
	{
		char prefix_arg[101];

		while( bytes_read = 0 ,
		       result = sscanf( str, " %*1[:;] %100[^:; ] %n", prefix_arg, &bytes_read ) ,
		       bytes_read && result == 1 )
		{
			str += bytes_read;

			if( !strcasecmp( prefix_arg, "undirected" ) )
			{
				is_undirected = true;
			}
			else if( strcmp( prefix_arg, "" ) )
			{
				ereport( ERROR,
				         (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
				          errmsg( "Invalid prefix argument for sparsegraph: \"%s\"", prefix_arg ) ) );
			}

			if( *str == ':' )
			{
				++str;
				break;
			}
		}
	}

	while( *str )
	{
		int32 label1;
		int32 label2;
		float8 weight;

		/* Either an edge or a (disconnected) node */
		if(( bytes_read = sscan_edge( str, &label1, &label2, &weight ) ))
		{
			mtx.addEdge( label1, label2, weight );
			if( is_undirected )
			{
				mtx.addEdge( label2, label1, weight );
			}
		}
		else if(( bytes_read = sscan_node( str, &label1 ) ))
		{
			mtx.addNode( label1 );
		}
		else
		{
			ereport( ERROR,
			         (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
			          errmsg( "Expected (LABEL) or (LABEL1, LABEL2, WEIGHT); got: \"%s\"", truncate_cstr( str, 32 ) ) ) );
		}

		str += bytes_read;

		if( *str == ',' )
		{
			str++;
		}
	}

	return mtx.newSparseMatrix();
}

const char *sparsegraph_export( struct sparsegraph *mtx )
{
	int32 i, j;
	ostringstream str;

	if( mtx->version > SPARSEGRAPH_CURRENT )
	{
		ereport( ERROR,
		         (errcode(ERRCODE_DATATYPE_MISMATCH),
		          errmsg( "sparsegraph code version (0x%08x) is older than data "
		                  "version (0x%08x)!", SPARSEGRAPH_CURRENT, mtx->version ) ) );
	}

	for( i = 0 ; i < mtx->num_nodes ; i++ )
	{
		label_t i_label = mtx->data[i].label;

		for( j = 0 ; j < mtx->data[i].num_edges ; j++ )
		{
			int32 idx = mtx->data[i].data_idx + j;
			label_t j_label = mtx->data[mtx->data[idx].node_idx].label;
			weight_t weight = mtx->data[idx].weight;

			str << "(" << i_label << "," << j_label << "," << weight << ")";
		}

		if( j == 0 )
		{
			/* There were no edges, so indicate the presence of a node */
			str << "(" << i_label << ")";
		}
	}

	return pstring( str.str() );
}
